/*-----------------------------------------------------------------------------
 * Copyright (c) 2013 - 2020 Arm Limited (or its affiliates). All
 * rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   1.Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   2.Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   3.Neither the name of Arm nor the names of its contributors may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * Name:    ADC_EA_LPC1788.c
 * Purpose: A/D Converter interface for Embedded Artists LPC1788-32 Developers Kit
 * Rev.:    1.02
   ---------------------------------------------------------------------------*/

#include "LPC177x_8x.h"
#include "PIN_LPC177x_8x.h"

#include "Board_ADC.h"

#define ADC_RESOLUTION        12U            /* Number of A/D converter bits  */

static volatile uint16_t AD_last;            /* Last converted value          */
static volatile uint8_t  AD_done;            /* AD conversion done flag       */

/* Interrupt handler prototype */
void ADC_IRQHandler (void);

/* ADC pins:
   - AD0.2: P0_25
*/

static const PIN ADC_PIN[] = {
  {0U, 25U},
};

/**
  \fn          int32_t ADC_Initialize (void)
  \brief       Initialize Analog-to-Digital Converter
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t ADC_Initialize (void) {

  LPC_SC->PCONP |= ((1U << 12) | (1U << 15));  /* enable power to ADC & IOCON   */

  PIN_Configure (ADC_PIN[0].Portnum, ADC_PIN[0].Pinnum, 1 | IOCON_ANALOG_MODE);
  
  LPC_ADC->CR      =  ( 1U <<  2) |     /* select AD0.2 pin                   */
                      ( 4U <<  8) |     /* ADC clock is 60MHz/5               */
                      ( 1U << 21);      /* enable ADC                         */

  LPC_ADC->INTEN   =  ( 1U <<  8);      /* global ADC enable interrupt        */

  NVIC_EnableIRQ(ADC_IRQn);             /* enable ADC Interrupt               */

  return 0;
}

/**
  \fn          int32_t ADC_Uninitialize (void)
  \brief       De-initialize Analog-to-Digital Converter
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t ADC_Uninitialize (void) {

  NVIC_DisableIRQ (ADC_IRQn);           /* Disable ADC Interrupt              */
  LPC_ADC->INTEN &= ~( 1U <<  8);       /* Disable global ADC enable interrupt */

  PIN_Configure (ADC_PIN[0].Portnum, ADC_PIN[0].Pinnum, 0U |\
      IOCON_DIGITIAL_MODE | IOCON_INVERT_INPUT | IOCON_MODE_PULLUP);
  
  LPC_SC->PCONP &= ~(1U << 12);         /* Disable ADC clock                  */

  return 0;
}

/**
  \fn          int32_t ADC_StartConversion (void)
  \brief       Start conversion
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t ADC_StartConversion (void) {

  LPC_ADC->CR &= ~( 7U << 24);          /* stop conversion                    */
  LPC_ADC->CR |=  ( 1U << 24);          /* start conversion                   */

  return 0;
}

/**
  \fn          int32_t ADC_ConversionDone (void)
  \brief       Check if conversion finished
  \returns
   - \b  0: conversion finished
   - \b -1: conversion in progress
*/
int32_t ADC_ConversionDone (void) {
  return (AD_done ? 0 : -1);
}

/**
  \fn          int32_t ADC_GetValue (void)
  \brief       Get converted value
  \returns
   - <b> >=0</b>: converted value
   - \b -1: conversion in progress or failed
*/
int32_t ADC_GetValue (void) {

  if (AD_done) {
    AD_done = 0U;
    return AD_last;
  }
  return -1;
}

/**
  \fn          uint32_t ADC_GetResolution (void)
  \brief       Get resolution of Analog-to-Digital Converter
  \returns     Resolution (in bits)
*/
uint32_t ADC_GetResolution (void) {
  return ADC_RESOLUTION;
}

/**
  \fn          void ADC_IRQHandler (void)
  \brief       Analog-to-Digital Converter Interrupt Handler
*/
void ADC_IRQHandler(void) {
  volatile uint32_t adstat;

  adstat = LPC_ADC->STAT;               /* Read ADC clears interrupt          */

  AD_last = (LPC_ADC->GDR >> 4) & 0xFFFU; /* Store converted value            */

  AD_done = 1U;
}
