/*-----------------------------------------------------------------------------
 * Copyright (c) 2013 - 2019 Arm Limited (or its affiliates). All
 * rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   1.Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   2.Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   3.Neither the name of Arm nor the names of its contributors may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * Name:    GLCD_EA_LPC1788.c
 * Purpose: Graphic LCD interface for Embedded Artists LPC1788-32 Developer's Kit
 * Rev.:    1.3
   ---------------------------------------------------------------------------*/

#include "LPC177x_8x.h"
#include "PIN_LPC177x_8x.h"
#include "GPIO_LPC17xx.h"
#include "GLCD_Config.h"
#include "Board_GLCD.h"
#include "RTE_Components.h"
#if defined (RTE_EA_LCD_TFT_320x240)
  #include "Driver_SPI.h"
#endif
#if defined (RTE_EA_LCD_TFT_480x272) || defined (RTE_EA_LCD_TFT_800x480)
  #include "Driver_I2C.h"
#endif

/* GLCD pins:
   - LCD_VD19  : P4_1
   - LCD_VD3   : P4_2
   - LCD_VSYNC : P4_5
   - LCD_EN    : P4_6
   - LCD_DOTCLK: P4_7
   - LCD_VD11  : P4_9
   - LCD_VD10  : P4_10
   - LCD_BL_EN : P7_0  = GPIO3[8]
   - LCD_HSYNC : P7_6
   - LCD_VD12  : P8_3
   - LCD_VD7   : P8_4
   - LCD_VD6   : P8_5
   - LCD_VD5   : P8_6
   - LCD_VD4   : P8_7
   - LCD_VD23  : PB_0
   - LCD_VD22  : PB_1
   - LCD_VD21  : PB_2
   - LCD_VD20  : PB_3
   - LCD_VD15  : PB_4
   - LCD_VD14  : PB_5
   - LCD_VD13  : PB_6             */

// LCD configuration
#if defined (RTE_EA_LCD_TFT_320x240)
  #define HBP_TRULY_3_2   28U        // Horizontal back porch in clocks
  #define HFP_TRULY_3_2   10U        // Horizontal front porch in clocks
  #define HSW_TRULY_3_2   2U         // HSYNC pulse width in clocks
  #define PPL_TRULY_3_2   240U       // Pixels per line
  #define VBP_TRULY_3_2   2U         // Vertical back porch in clocks
  #define VFP_TRULY_3_2   1U         // Vertical front porch in clocks
  #define VSW_TRULY_3_2   2U         // VSYNC pulse width in clocks
  #define LPP_TRULY_3_2   320U       // Lines per panel
  #define IOE_TRULY_3_2   0U         // Invert output enable, 1 = invert
  #define IPC_TRULY_3_2   1U         // Invert panel clock, 1 = invert
  #define IHS_TRULY_3_2   1U         // Invert HSYNC, 1 = invert
  #define IVS_TRULY_3_2   1U         // Invert VSYNC, 1 = invert
  #define ACB_TRULY_3_2   1U         // AC bias frequency in clocks (not used)
  #define BPP_TRULY_3_2   LCDC_BPP_16_565       // Bits per pixel b110 = 16 bpp 5:6:5 mode
  #define CLK_TRULY_3_2   8200000U   // Optimal clock rate (Hz) between 1-8.22 MHz according to SSD1289 datasheet
  #define LCD_TRULY_3_2   0U         // Panel type; 0: LCD TFT panel
  #define DUAL_TRULY_3_2  0U         // No dual panel
#elif defined (RTE_EA_LCD_TFT_480x272)
  #define HBP_TRULY_3_2   40U        // Horizontal back porch in clocks
  #define HFP_TRULY_3_2   5U         // Horizontal front porch in clocks
  #define HSW_TRULY_3_2   2U         // HSYNC pulse width in clocks
  #define PPL_TRULY_3_2   480U       // Pixels per line
  #define VBP_TRULY_3_2   8U         // Vertical back porch in clocks
  #define VFP_TRULY_3_2   8U         // Vertical front porch in clocks
  #define VSW_TRULY_3_2   2U         // VSYNC pulse width in clocks
  #define LPP_TRULY_3_2   272U       // Lines per panel
  #define IOE_TRULY_3_2   0U         // Invert output enable, 1 = invert
  #define IPC_TRULY_3_2   0U         // Invert panel clock, 1 = invert
  #define IHS_TRULY_3_2   1U         // Invert HSYNC, 1 = invert
  #define IVS_TRULY_3_2   1U         // Invert VSYNC, 1 = invert
  #define ACB_TRULY_3_2   1U         // AC bias frequency in clocks (not used)
  #define BPP_TRULY_3_2   LCDC_BPP_16_565       // Bits per pixel b110 = 16 bpp 5:6:5 mode
  #define CLK_TRULY_3_2   9000000U   // Optimal clock rate (Hz) between 1-8.22 MHz according to SSD1289 datasheet
  #define LCD_TRULY_3_2   LCDC_TFT   // Panel type; 0: LCD TFT panel
  #define DUAL_TRULY_3_2  0U         // No dual panel
#elif defined (RTE_EA_LCD_TFT_800x480)
  #define HBP_TRULY_3_2   45U        // Horizontal back porch in clocks
  #define HFP_TRULY_3_2   17U        // Horizontal front porch in clocks
  #define HSW_TRULY_3_2   2U         // HSYNC pulse width in clocks
  #define PPL_TRULY_3_2   800U       // Pixels per line
  #define VBP_TRULY_3_2   22U         // Vertical back porch in clocks
  #define VFP_TRULY_3_2   22U        // Vertical front porch in clocks
  #define VSW_TRULY_3_2   2U         // VSYNC pulse width in clocks
  #define LPP_TRULY_3_2   480U       // Lines per panel
  #define IOE_TRULY_3_2   0U         // Invert output enable, 1 = invert
  #define IPC_TRULY_3_2   0U         // Invert panel clock, 1 = invert
  #define IHS_TRULY_3_2   1U         // Invert HSYNC, 1 = invert
  #define IVS_TRULY_3_2   1U         // Invert VSYNC, 1 = invert
  #define ACB_TRULY_3_2   1U         // AC bias frequency in clocks (not used)
  #define BPP_TRULY_3_2   LCDC_BPP_16_565       // Bits per pixel b110 = 16 bpp 5:6:5 mode
  #define CLK_TRULY_3_2   36000000U  // Optimal clock rate (Hz) between 1-8.22 MHz according to SSD1289 datasheet
  #define LCD_TRULY_3_2   LCDC_TFT   // Panel type; 0: LCD TFT panel
  #define DUAL_TRULY_3_2  0U         // No dual panel
#endif


// LCDC
#define LCDC_BPP_16_565    6U    // Bits per pixel b110 = 16 bpp 5:6:5 mode
#define LCDC_TFT           1U    // TFT value for LCDC

#define BGR                1U    // Swap red and blue

#define LCD_BACKLIGHT_PIN  18U   // P1.18
#define LCD_REG_PIN        19U   // P0.19


#define MATRIX_ARB  (*(__IO uint32_t *)(0x400FC188))

// PCA9532 (I2C) on the display BCB
#define PCA9532_I2C_ADDR         0x64U
#define PCA9532_DISP_3V3_EN_BIT  0U
#define PCA9532_DISP_5V_EN_BIT   1U
#define PCA9532_DISP_EN_BIT      4U
#define PCA9532_DISP_BL_BIT      8U

// PCA9532 (I2C) Register addresses
#define INPUT0  0U
#define INPUT1  1U
#define PSC0    2U
#define PWM0    3U
#define PSC1    4U
#define PWM1    5U
#define LS0     6U
#define LS1     7U
#define LS2     8U
#define LS3     9U

#define I2C_PCA9532_HIGHZ  0U
#define I2C_PCA9532_LOW    1U
#define I2C_PCA9532_PWM0   2U
#define I2C_PCA9532_PWM1   3U

// Types
typedef struct {
  uint8_t  HBP;          // Horizontal back porch in clocks
  uint8_t  HFP;          // Horizontal front porch in clocks
  uint8_t  HSW;          // HSYNC pulse width in clocks
  uint8_t  Reserved0;    // Reserved
  uint16_t PPL;          // Pixels per line
  uint8_t  VBP;          // Vertical back porch in clocks
  uint8_t  VFP;          // Vertical front porch in clocks
  uint8_t  VSW;          // VSYNC pulse width in clocks
  uint8_t  Reserved1;    // Reserved
  uint16_t LPP;          // Lines per panel
  uint8_t  IOE;          // Invert output enable, 1 = invert
  uint8_t  IPC;          // Invert panel clock, 1 = invert
  uint8_t  IHS;          // Invert HSYNC, 1 = invert
  uint8_t  IVS;          // Invert VSYNC, 1 = invert
  uint8_t  ACB;          // AC bias frequency in clocks (not used)
  uint8_t  BPP;          // Maximum bits per pixel the display supports
  uint8_t  Reserved2[2]; // Reserved
  uint32_t Clk;          // Optimal clock rate (Hz)
  uint8_t  LCD;          // LCD panel type
  uint8_t  Dual;         // Dual panel, 1 = dual panel display
  uint8_t  Reserved3[2]; // Reserved
} LCD_PARAM;


/* SPI Interface: SSP0
 
   PINS: 
   - CS  = PF_1 (GPIO7[16])
   - SCK = PF_0 (SCK0)
   - SDO = PF_2 (MISO0)
   - SDI = PF_3 (MOSI0)
*/

#ifndef PCA9532_I2C_PORT
#define PCA9532_I2C_PORT    0
#endif
#ifdef RTE_EA_LCD_TFT_320x240
/* SPI Driver */
extern ARM_DRIVER_SPI         Driver_SPI0;
#define ptrSPI              (&Driver_SPI0)
#endif

#if defined (RTE_EA_LCD_TFT_480x272) || defined (RTE_EA_LCD_TFT_800x480)
/* I2C Driver */
#define _I2C_Driver_(n)  Driver_I2C##n
#define  I2C_Driver_(n) _I2C_Driver_(n)
extern ARM_DRIVER_I2C    I2C_Driver_(PCA9532_I2C_PORT);
#define ptrI2C         (&I2C_Driver_(PCA9532_I2C_PORT))
#endif

#ifndef VRAM_BASE_ADDR
#define VRAM_BASE_ADDR       0xA0000000U
#endif
#ifndef      __MEMORY_AT
  #if     (defined (__CC_ARM))
    #define  __MEMORY_AT(x)     __attribute__((at(x)))
  #elif   (defined (__ARMCC_VERSION) && (__ARMCC_VERSION >= 6010050))
    #define  __MEMORY_AT__(x)   __attribute__((section(".bss.ARM.__at_"#x)))
    #define  __MEMORY_AT(x)     __MEMORY_AT__(x)
  #else
    #define  __MEMORY_AT(x)
    #warning Position memory containing __MEMORY_AT macro at absolute address!
  #endif
#endif


static const LCD_PARAM  _LcdParams = {
  HBP_TRULY_3_2,
  HFP_TRULY_3_2,
  HSW_TRULY_3_2,
  0U,
  PPL_TRULY_3_2,
  VBP_TRULY_3_2,
  VFP_TRULY_3_2,
  VSW_TRULY_3_2,
  0U,
  LPP_TRULY_3_2,
  IOE_TRULY_3_2,
  IPC_TRULY_3_2,
  IHS_TRULY_3_2,
  IVS_TRULY_3_2,
  ACB_TRULY_3_2,
  BPP_TRULY_3_2,
  {0U, 0U},
  CLK_TRULY_3_2,
  LCD_TRULY_3_2,
  DUAL_TRULY_3_2,
  {0U, 0U}
};

/*---------------------------- Global variables ------------------------------*/

static uint16_t   frame_buf[GLCD_WIDTH*GLCD_HEIGHT] __MEMORY_AT(VRAM_BASE_ADDR);
static uint16_t   foreground_color = GLCD_COLOR_BLACK;
static uint16_t   background_color = GLCD_COLOR_WHITE;
static GLCD_FONT *active_font      = NULL;

/************************ Local auxiliary functions ***************************/

/**
  \fn          void delay_ms (int ms)
  \brief       Delay execution for a specified number of milliseconds
  \param[in]   ms     number of while loop iterations
*/
static void delay_ms (int ms) {
  ms *= (SystemCoreClock/10000);
  while (ms--) { __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); }
}

/**
  \fn          uint32_t GetPanelClkDivisor(uint32_t clk)
  \brief       Get panel clock divisor
  \param[in]   clk  LCD panel clock value
  \returns     Devisor value
*/
static uint32_t GetPanelClkDivisor(uint32_t clk) {
  uint32_t divider;
  uint32_t r;

  divider = 1;
  while (((SystemCoreClock / divider) > clk) && (divider <= 0x3FU)) {
    divider++;
  }
  if (divider <= 1) {
    r = (1U << 26);        // Skip divider logic if clock divider is 1
  } else {
    // Use found divider
    divider -= 2;
    r = 0U | (((divider >> 0) & 0x1FU) | (((divider >> 5) & 0x1FU) << 27));
  }
  return r;
}

/**
  \fn          void LCDC_Initialize (void)
  \brief       Initialize LCD Controller
*/
static void LCDC_Initialize(void) {
  uint32_t i;

  LPC_SC->PCONP |=  (1UL << 0);  // Power the LCDC
  
  LPC_LCD->CTRL &= ~(1UL << 0);  // Disable the LCDC
  LPC_LCD->TIMH  = 0             // Configure horizontal axis
                   | ((((uint32_t)_LcdParams.PPL / 16) - 1) <<  2)
                   |  (((uint32_t)_LcdParams.HSW - 1)       <<  8)
                   |  (((uint32_t)_LcdParams.HFP - 1)       << 16)
                   |  (((uint32_t)_LcdParams.HBP - 1)       << 24)
                   ;
  LPC_LCD->TIMV  = 0             // Configure vertical axis
                   | (((uint32_t)_LcdParams.LPP - 1) <<  0)
                   | (((uint32_t)_LcdParams.VSW - 1) << 10)
                   | (((uint32_t)_LcdParams.VFP)     << 16)
                   | (((uint32_t)_LcdParams.VBP)     << 24)
                   ;
  LPC_LCD->POL   = 0             // Configure clock and signal polarity
                   | (GetPanelClkDivisor(_LcdParams.Clk) <<  0)
                   | (((uint32_t)_LcdParams.ACB - 1)         <<  6)
                   | (((uint32_t)_LcdParams.IVS)             << 11)
                   | (((uint32_t)_LcdParams.IHS)             << 12)
                   | (((uint32_t)_LcdParams.IPC)             << 13)
                   | (((uint32_t)_LcdParams.PPL - 1)         << 16)
                   ;
  LPC_LCD->CTRL  = 0             // Configure operating mode and panel parameters
                   | ((uint32_t)_LcdParams.BPP << 1)
                   | ((uint32_t)BGR            << 8)
                   | ((uint32_t)_LcdParams.LCD << 5)
                   ;
  for (i = 0U; i < 128U; i++) {
    LPC_LCD->PAL[i] = 0;  // Clear the color palette with black
  }
  LPC_SC->LCD_CFG = 0x0;  // No panel clock prescaler

  // Enable LCDC
  LPC_LCD->UPBASE  = VRAM_BASE_ADDR;
  LPC_LCD->CTRL   |= (1U <<  0);                 // Enable LCD signals
  LPC_LCD->CTRL   |= (1U << 11);                 // Enable LCD power
  LPC_GPIO1->SET  |= (1U << LCD_BACKLIGHT_PIN);  // Set backlight to on
}

#ifdef RTE_EA_LCD_TFT_320x240
/**
  \fn          void wr_reg(uint16_t addr, uint16_t value)
  \brief       Write to display register
  \param[in]   addr  value
  \param[in]   value value
*/

static void wr_reg(uint16_t addr, uint16_t value) {
  uint8_t data[2];

  GPIO_PinWrite(1, LCD_REG_PIN, 0U);
  
  data[0] = 0U;
  data[1] = addr & 0xFFU;
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  ptrSPI->Send   (data, 2U);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);

  GPIO_PinWrite(1U, LCD_REG_PIN, 1U);

  data[0] = value >> 8;
  data[1] = value & 0xFF;
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  ptrSPI->Send   (data, 2);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);

  GPIO_PinWrite(1U, LCD_REG_PIN, 0U);
  
  data[0] = 0U;
  data[1] = 0x22U;
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  ptrSPI->Send   (data, 2U);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);
}
#endif

/*********************************************************************
*
*       Local functions for LCDs other than Truly
*
**********************************************************************
*/

#if defined (RTE_EA_LCD_TFT_480x272) || defined (RTE_EA_LCD_TFT_800x480)
/**
  \fn          int8_t SetPinFuncSelect(uint8_t pin, uint8_t state)
  \brief       Sets pin function select for one pin
  \param[in]   pin    pin number
  \param[in]   state  pin state
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/

static int8_t SetPinFuncSelect(uint8_t pin, uint8_t state) {
  uint8_t data[2];

  if        (pin < 4) {
    data[0]  = LS0;
  } else if (pin < 8) {
    data[0]  = LS1;
    pin     -= 4;
  } else if (pin < 12) {
    data[0]  = LS2;
    pin     -= 8;
  } else if (pin < 16) {
    data[0]  = LS3;
    pin     -= 12;
  } else {
    return -1;  // Error, invalid pin
  }
  pin *= 2;

  ptrI2C->MasterTransmit (PCA9532_I2C_ADDR, &data[0], 1U, true);
  while (ptrI2C->GetStatus().busy);
  ptrI2C->MasterReceive (PCA9532_I2C_ADDR, &data[1], 1U, false);
  while (ptrI2C->GetStatus().busy);
  
  data[1] &= ~(0x3U   << pin);
  data[1] |=  (state << pin);
  ptrI2C->MasterTransmit (PCA9532_I2C_ADDR, data, 2U, false);
  while (ptrI2C->GetStatus().busy);

  return 0;
}

/*********************************************************************
*
*       I2C_PCA9532_SetPwm
*
* Function description:
*   Sets the timing for PWM0 or PWM1 and assigns the pin to the
*   according PWM.
*
* Return value:
*      0: O.K.
*   != 0: Error
*/
/**
  \fn          int8_t SetPwm(uint8_t pin, uint8_t state)
  \brief       Sets the timing for PWM0 or PWM1
  \param[in]   pin      pin number
  \param[in]   pwm_num  pwm number
  \param[in]   pwm_val  pwm value
  \param[in]   psc      psc value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/

static int8_t SetPwm (uint8_t pin, uint8_t pwm_num, uint8_t pwm_val, uint8_t psc) {
  uint8_t data[2];

  if (pwm_num > 1) return 1;

  data[0] = (pwm_num) ? PSC1 : PSC0;
  data[1] = psc;

  ptrI2C->MasterTransmit (PCA9532_I2C_ADDR, data, 2U, false);
  while (ptrI2C->GetStatus().busy);

  data[0] = (pwm_num) ? PWM1 : PWM0;
  data[1] = pwm_val;

  ptrI2C->MasterTransmit (PCA9532_I2C_ADDR, data, 2U, false);
  while (ptrI2C->GetStatus().busy);

  SetPinFuncSelect(pin, I2C_PCA9532_PWM0 + pwm_num);
  return 0;
}
#endif

/************************ Module functions ************************************/

/**
  \fn          int32_t GLCD_Initialize (void)
  \brief       Initialize Graphic LCD
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_Initialize (void) {

  SystemCoreClockUpdate();

#ifdef RTE_EA_LCD_TFT_320x240
  ptrSPI->Initialize  (NULL);
  ptrSPI->PowerControl(ARM_POWER_FULL);
  ptrSPI->Control     (ARM_SPI_MODE_MASTER | ARM_SPI_CPOL1_CPHA1 | ARM_SPI_MSB_LSB | ARM_SPI_SS_MASTER_SW | ARM_SPI_DATA_BITS(8), 10000000U);
#elif defined (RTE_EA_LCD_TFT_480x272) || defined (RTE_EA_LCD_TFT_800x480)
  if (ptrI2C->Initialize (NULL) == ARM_DRIVER_OK) 
  {
    ptrI2C->PowerControl (ARM_POWER_FULL);
    ptrI2C->Control      (ARM_I2C_BUS_SPEED, ARM_I2C_BUS_SPEED_FAST);
    ptrI2C->Control      (ARM_I2C_BUS_CLEAR, 0U);
  }
#endif  

  // Setup BUS priorities
  MATRIX_ARB = 0            // Set AHB Matrix priorities [0..3] with 3 being highest priority
               | (1U <<  0) // PRI_ICODE : I-Code bus priority. Should be lower than PRI_DCODE for proper operation.
               | (2U <<  2) // PRI_DCODE : D-Code bus priority.
               | (0U <<  4) // PRI_SYS   : System bus priority.
               | (0U <<  6) // PRI_GPDMA : General Purpose DMA controller priority.
               | (0U <<  8) // PRI_ETH   : Ethernet: DMA priority.
               | (3U << 10) // PRI_LCD   : LCD DMA priority.
               | (0U << 12) // PRI_USB   : USB DMA priority.
               | (1U << 16) // ROM_LAT   : ROM latency select.
               ;

  /* Enable GPIO Clock */
  GPIO_PortClock (1U);

  /* Configure pins */
  PIN_Configure (0U,   4U,  7U);  // LCD_VD_0
  PIN_Configure (0U,   5U,  7U);  // LCD_VD_1
  PIN_Configure (0U,   6U,  7U);  // LCD_VD_8
  PIN_Configure (0U,   7U,  7U);  // LCD_VD_9
  PIN_Configure (0U,   8U,  7U);  // LCD_VD_16
  PIN_Configure (0U,   9U,  7U);  // LCD_VD_17

  PIN_Configure (1U,  20U,  7U);  // LCD_VD_10
  PIN_Configure (1U,  21U,  7U);  // LCD_VD_11
  PIN_Configure (1U,  22U,  7U);  // LCD_VD_12
  PIN_Configure (1U,  23U,  7U);  // LCD_VD_13
  PIN_Configure (1U,  24U,  7U);  // LCD_VD_14
  PIN_Configure (1U,  25U,  7U);  // LCD_VD_15
  PIN_Configure (1U,  26U,  7U);  // LCD_VD_20
  PIN_Configure (1U,  27U,  7U);  // LCD_VD_21
  PIN_Configure (1U,  28U,  7U);  // LCD_VD_22
  PIN_Configure (1U,  29U,  7U);  // LCD_VD_23
  PIN_Configure (2U,   0U,  7U);  // LCD_PWR
  PIN_Configure (2U,   1U,  7U);  // LCD_LE
  PIN_Configure (2U,   2U,  7U);  // LCD_DCLK
  PIN_Configure (2U,   3U,  7U);  // LCD_FP
  PIN_Configure (2U,   4U,  7U);  // LCD_ENAB_M
  PIN_Configure (2U,   5U,  7U);  // LCD_LP
  PIN_Configure (2U,   6U,  7U);  // LCD_VD_4
  PIN_Configure (2U,   7U,  7U);  // LCD_VD_5
  PIN_Configure (2U,   8U,  7U);  // LCD_VD_6
  PIN_Configure (2U,   9U,  7U);  // LCD_VD_7

  PIN_Configure (2U,  11U,  7U);  // LCD_CLKIN

#ifdef RTE_EA_LCD_TFT_320x240
  PIN_Configure (2U,  12U,  7U);  // LCD_VD_18
  PIN_Configure (4U,  29U,  7U);  // LCD_VD_3
#elif defined (RTE_EA_LCD_TFT_480x272) || defined (RTE_EA_LCD_TFT_800x480)
  PIN_Configure (2U,  12U,  5U);  // P2.12 is marked as LCD_VD_18 on base board, but shall carry the LCD_VD_3 signal for displays other than TRULY
#endif

  PIN_Configure (2U,  13U,  7U);  // LCD_VD_19
  PIN_Configure (4U,  28U,  7U);  // LCD_VD_2

  /* Configure backlight pin GPIO mode */
  GPIO_SetDir (1U, LCD_BACKLIGHT_PIN, GPIO_DIR_OUTPUT);

#ifdef RTE_EA_LCD_TFT_320x240
  // Set LCD power pin to output high
  GPIO_SetDir  (2U, 0U, GPIO_DIR_OUTPUT);
  GPIO_PinWrite(2U, 0U, 1U);
  
  // Set display CMD/DATA register select pin to output high
  GPIO_SetDir(1U, LCD_REG_PIN, GPIO_DIR_OUTPUT);

  // Display init sequence
  wr_reg(0x00U,0x0001U);
  delay_ms(15U * 2U);
  wr_reg(0x03U,0x6E3EU);  // AAAC
  wr_reg(0x0CU,0x0007U);  // 0002
  wr_reg(0x0DU,0x000EU);  // 000A
  wr_reg(0x0EU,0x2C00U);  // 2C00
  wr_reg(0x1EU,0x00AEU);  // 00B8
  delay_ms(15U * 2U);

  wr_reg(0x07U,0x0021U);
  delay_ms(50U * 2U);
  wr_reg(0x07U,0x0023U);
  delay_ms(50U * 2U);
  wr_reg(0x07U,0x0033U);
  delay_ms(50U * 2U);

  wr_reg(0x01U,0x2B3FU);
  wr_reg(0x02U,0x0600U);
  wr_reg(0x10U,0x0000U);
  delay_ms(15U * 2U);
  wr_reg(0x11U,0xC5B0U);  // 60B0: RGB I/R
  delay_ms(20U * 2U);
  wr_reg(0x15U,0x00D0U);

  wr_reg(0x05U,0x0000U);
  wr_reg(0x06U,0x0000U);
  wr_reg(0x16U,0xEF1CU);
  wr_reg(0x17U,0x0003U);
  wr_reg(0x07U,0x0233U);
  wr_reg(0x0BU,0x5310U);
  wr_reg(0x0FU,0x0000U);
  wr_reg(0x25U,0xE000U);
  delay_ms(20U * 2U);

  wr_reg(0x41U,0x0000U);
  wr_reg(0x42U,0x0000U);
  wr_reg(0x48U,0x0000U);
  wr_reg(0x49U,0x013FU);
  wr_reg(0x44U,0xEF00U);
  wr_reg(0x45U,0x0000U);
  wr_reg(0x46U,0x013FU);
  wr_reg(0x4AU,0x0000U);
  wr_reg(0x4BU,0x0000U);
  delay_ms(20U * 2U);

  wr_reg(0x30U,0x0707U);
  wr_reg(0x31U,0x0600U);  // 0704
  wr_reg(0x32U,0x0005U);  // 0204
  wr_reg(0x33U,0x0402U);  // 0201
  wr_reg(0x34U,0x0203U);
  wr_reg(0x35U,0x0204U);
  wr_reg(0x36U,0x0204U);
  wr_reg(0x37U,0x0401U);  // 0502
  wr_reg(0x3AU,0x0302U);
  wr_reg(0x3BU,0x0500U);
  delay_ms(20U * 2U);
  wr_reg(0x22U,0x0000U);

  delay_ms(20U * 2U);

  LCDC_Initialize();
#elif defined (RTE_EA_LCD_TFT_480x272)
  SetPinFuncSelect(PCA9532_DISP_EN_BIT, 1U);
  SetPwm(PCA9532_DISP_BL_BIT, 0U, 255U, 0U);
  SetPinFuncSelect(PCA9532_DISP_3V3_EN_BIT, 0);

  delay_ms (100U);

  SetPinFuncSelect(PCA9532_DISP_3V3_EN_BIT, 1U);

  delay_ms (100U);
  SetPinFuncSelect(PCA9532_DISP_EN_BIT, 0U);

  delay_ms (10U);

  LCDC_Initialize();
  SetPwm(PCA9532_DISP_BL_BIT, 0U, 0U, 0U);
#elif defined (RTE_EA_LCD_TFT_800x480)
  SetPwm (PCA9532_DISP_BL_BIT, 0U, 255U, 0U);
  SetPinFuncSelect(PCA9532_DISP_3V3_EN_BIT, 1U);

  delay_ms (50U);

  LCDC_Initialize();

  delay_ms (200U);

  SetPinFuncSelect(PCA9532_DISP_5V_EN_BIT, 1U);
  SetPwm (PCA9532_DISP_BL_BIT, 0U, 0U, 0U);
#endif

  GPIO_PinWrite(1U, LCD_BACKLIGHT_PIN, 1U);   // Set backlight on
  return 0;
}

/**
  \fn          int32_t GLCD_Uninitialize (void)
  \brief       De-initialize Graphic LCD
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_Uninitialize (void) {

  GPIO_PinWrite(1U, LCD_BACKLIGHT_PIN, 0U);
  /* Unconfigure pins */
  PIN_Configure (0U,   4U,  0U);  // LCD_VD_0
  PIN_Configure (0U,   5U,  0U);  // LCD_VD_1
  PIN_Configure (0U,   6U,  0U);  // LCD_VD_8
  PIN_Configure (0U,   7U,  0U);  // LCD_VD_9
  PIN_Configure (0U,   8U,  0U);  // LCD_VD_16
  PIN_Configure (0U,   9U,  0U);  // LCD_VD_17

  PIN_Configure (1U,  20U,  0U);  // LCD_VD_10
  PIN_Configure (1U,  21U,  0U);  // LCD_VD_11
  PIN_Configure (1U,  22U,  0U);  // LCD_VD_12
  PIN_Configure (1U,  23U,  0U);  // LCD_VD_13
  PIN_Configure (1U,  24U,  0U);  // LCD_VD_14
  PIN_Configure (1U,  25U,  0U);  // LCD_VD_15
  PIN_Configure (1U,  26U,  0U);  // LCD_VD_20
  PIN_Configure (1U,  27U,  0U);  // LCD_VD_21
  PIN_Configure (1U,  28U,  0U);  // LCD_VD_22
  PIN_Configure (1U,  29U,  0U);  // LCD_VD_23
  PIN_Configure (2U,   0U,  0U);  // LCD_PWR
  PIN_Configure (2U,   1U,  0U);  // LCD_LE
  PIN_Configure (2U,   2U,  0U);  // LCD_DCLK
  PIN_Configure (2U,   3U,  0U);  // LCD_FP
  PIN_Configure (2U,   4U,  0U);  // LCD_ENAB_M
  PIN_Configure (2U,   5U,  0U);  // LCD_LP
  PIN_Configure (2U,   6U,  0U);  // LCD_VD_4
  PIN_Configure (2U,   7U,  0U);  // LCD_VD_5
  PIN_Configure (2U,   8U,  0U);  // LCD_VD_6
  PIN_Configure (2U,   9U,  0U);  // LCD_VD_7

  PIN_Configure (2U,  11U,  0U);  // LCD_CLKIN

#ifdef RTE_EA_LCD_TFT_320x240
  PIN_Configure (2U,  12U,  0U);  // LCD_VD_18
  PIN_Configure (4U,  29U,  0U);  // LCD_VD_3
#elif defined (RTE_EA_LCD_TFT_480x272) || defined (RTE_EA_LCD_TFT_800x480)
  PIN_Configure (2U,  12U,  0U);  // P2.12 is marked as LCD_VD_18 on base board, but shall carry the LCD_VD_3 signal for displays other than TRULY
#endif

  PIN_Configure (2U,  13U,  0U);  // LCD_VD_19
  PIN_Configure (4U,  28U,  0U);  // LCD_VD_2
  
  return 0;
}

/**
  \fn          int32_t GLCD_SetForegroundColor (uint32_t color)
  \brief       Set foreground color
  \param[in]   color  Color value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetForegroundColor (uint32_t color) {
  foreground_color = color & 0xFFFFU;
  return 0;
}

/**
  \fn          int32_t GLCD_SetBackgroundColor (uint32_t color)
  \brief       Set background color
  \param[in]   color  Color value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetBackgroundColor (uint32_t color) {
  background_color = color & 0xFFFFU;
  return 0;
}

/**
  \fn          int32_t GLCD_ClearScreen (void)
  \brief       Clear screen (with active background color)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_ClearScreen (void) {
  uint32_t i;
  uint16_t *ptr_frame_buf;

  ptr_frame_buf = frame_buf;
  for (i = 0U; i < (GLCD_WIDTH * GLCD_HEIGHT); i++) {
    *ptr_frame_buf++ = background_color;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_SetFont (GLCD_FONT *font)
  \brief       Set active font
  \param[in]   font   Pointer to font structure
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetFont (GLCD_FONT *font) {
  active_font = font;
  return 0;
}

/**
  \fn          int32_t GLCD_DrawPixel (uint32_t x, uint32_t y)
  \brief       Draw pixel (in active foreground color)
  \param[in]   x  x position in pixels (0 = left corner)
  \param[in]   y  y position in pixels (0 = upper corner)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawPixel (uint32_t x, uint32_t y) {
  frame_buf[GLCD_SWAP_XY * (x * GLCD_HEIGHT + y) + (!GLCD_SWAP_XY) * (y * GLCD_WIDTH + x)] = foreground_color;
  return 0;
}

/**
  \fn          int32_t GLCD_DrawHLine (uint32_t x, uint32_t y, uint32_t length)
  \brief       Draw horizontal line (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   length Line length
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawHLine (uint32_t x, uint32_t y, uint32_t length) {
  uint32_t dot;

  dot = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y) + (!GLCD_SWAP_XY) * (y * GLCD_WIDTH + x);
  while (length--) { 
    frame_buf[dot] = foreground_color;
    dot += GLCD_SWAP_XY * GLCD_HEIGHT + (!GLCD_SWAP_XY);
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawVLine (uint32_t x, uint32_t y, uint32_t length)
  \brief       Draw vertical line (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   length Line length in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawVLine (uint32_t x, uint32_t y, uint32_t length) {
  uint32_t dot;

  dot = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y) + (!GLCD_SWAP_XY) * (y * GLCD_WIDTH + x);
  while (length--) { 
    frame_buf[dot] = foreground_color;
    dot += (!GLCD_SWAP_XY) * GLCD_WIDTH + (GLCD_SWAP_XY);
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawRectangle (uint32_t x, uint32_t y, uint32_t width, uint32_t height)
  \brief       Draw rectangle (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Rectangle width in pixels
  \param[in]   height Rectangle height in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawRectangle (uint32_t x, uint32_t y, uint32_t width, uint32_t height) {

  GLCD_DrawHLine (x,         y,          width);
  GLCD_DrawHLine (x,         y + height, width);
  GLCD_DrawVLine (x,         y,          height);
  GLCD_DrawVLine (x + width, y,          height);

  return 0;
}

/**
  \fn          int32_t GLCD_DrawChar (uint32_t x, uint32_t y, int32_t ch)
  \brief       Draw character (in active foreground color)
  \param[in]   x      x position in pixels (0 = left corner)
  \param[in]   y      y position in pixels (0 = upper corner)
  \param[in]   ch     Character
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawChar (uint32_t x, uint32_t y, int32_t ch) {
  uint32_t i, j;
  uint32_t wb, dot;
  uint8_t *ptr_ch_bmp;

  if (active_font == NULL) return -1;

  ch        -= active_font->offset;
  wb         = (active_font->width + 7)/8;
  ptr_ch_bmp = (uint8_t *)(uint32_t)active_font->bitmap + ((uint32_t)ch * wb * active_font->height);
  dot        = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y) + (!GLCD_SWAP_XY) * (y * GLCD_WIDTH + x);

  for (i = 0U; i < active_font->height; i++) {
    for (j = 0U; j < active_font->width; j++) {
      frame_buf[dot] = (((*ptr_ch_bmp >> (j & 7)) & 1) ? foreground_color : background_color);
      dot += GLCD_SWAP_XY * GLCD_HEIGHT + (!GLCD_SWAP_XY);
      if (((j & 7) == 7) && (j != (active_font->width - 1))) ptr_ch_bmp++;
    }
    dot = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y + (i + 1)) + (!GLCD_SWAP_XY) * ((y + i + 1) * GLCD_WIDTH + x);
    ptr_ch_bmp++;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawString (uint32_t x, uint32_t y, const char *str)
  \brief       Draw string (in active foreground color)
  \param[in]   x      x position in pixels (0 = left corner)
  \param[in]   y      y position in pixels (0 = upper corner)
  \param[in]   str    Null-terminated String
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawString (uint32_t x, uint32_t y, const char *str) {

  while (*str) { GLCD_DrawChar(x, y, *str++); x += active_font->width; }

  return 0;
}

/**
  \fn          int32_t GLCD_DrawBargraph (uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint32_t val)
  \brief       Draw bargraph (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Full bargraph width in pixels
  \param[in]   height Full bargraph height in pixels
  \param[in]   val    Active bargraph value (maximum value is 100)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawBargraph (uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint32_t val) {
  uint32_t i, j;
  uint32_t dot;

  val = (val * width) / 100;            /* Scale value                        */
  dot = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y) + (!GLCD_SWAP_XY) * (y * GLCD_WIDTH + x);

  for (i = 0U; i < height; i++) {
    for (j = 0U; j < width; j++) {
      frame_buf[dot] = ((j >= val) ? background_color : foreground_color);
      dot += GLCD_SWAP_XY * GLCD_HEIGHT + (!GLCD_SWAP_XY);
    }
    dot = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y + (i + 1)) + (!GLCD_SWAP_XY) * ((y + i + 1) * GLCD_WIDTH + x);
  }
  return 0;
}

/**
  \fn          int32_t GLCD_DrawBitmap (uint32_t x, uint32_t y, uint32_t width, uint32_t height, const uint8_t *bitmap)
  \brief       Draw bitmap (bitmap from BMP file without header)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Bitmap width in pixels
  \param[in]   height Bitmap height in pixels
  \param[in]   bitmap Bitmap data
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawBitmap (uint32_t x, uint32_t y, uint32_t width, uint32_t height, const uint8_t *bitmap) {
  uint32_t  i, j;
  uint32_t  dot;
  uint16_t *ptr_bmp;

  ptr_bmp = (uint16_t *)(uint32_t)bitmap + (width * (height - 1));
  dot     = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y) + (!GLCD_SWAP_XY) * (y * GLCD_WIDTH + x);

  for (i = 0; i < height; i++) {
    for (j = 0; j < width; j++) {
      frame_buf[dot] = *ptr_bmp++;
      dot += GLCD_SWAP_XY * GLCD_HEIGHT + (!GLCD_SWAP_XY);
    }
    dot = GLCD_SWAP_XY * (x * GLCD_HEIGHT + y + (i + 1)) + (!GLCD_SWAP_XY) * ((y + i + 1) * GLCD_WIDTH + x);
    ptr_bmp -= 2 * width;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_VScroll (uint32_t dy)
  \brief       Scroll picture on display vertically
  \param[in]   dy  Scroll size in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_VScroll (uint32_t dy) {
  uint32_t i, dif, len;

  dif = GLCD_SWAP_XY * dy + (!GLCD_SWAP_XY) * (GLCD_WIDTH * dy);
  len = GLCD_WIDTH * GLCD_HEIGHT - dif;
  for (i = 0U; i < len; i++) {
    frame_buf[i] = frame_buf[i+dif];
  }
  for (; i < GLCD_WIDTH * GLCD_HEIGHT; i++) {
    frame_buf[i] = background_color;
  }
  return 0;
}

/**
  \fn          int32_t GLCD_FrameBufferAccess (bool enable)
  \brief       Enable or disable direct access to FrameBuffer
  \param[in]   enable Enable or disable request
                - true (1): enable
                - false(0): disable
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t  GLCD_FrameBufferAccess   (bool enable) {
  (void)enable;
  /* Direct data access is used with direct access to frame buffer            */
  return 0;
}

/**
  \fn          uint32_t GLCD_FrameBufferAddress (void)
  \brief       Retrieve FrameBuffer address
  \returns     FrameBuffer address
*/
uint32_t GLCD_FrameBufferAddress (void) {
  return ((uint32_t)frame_buf);
}
