/*-----------------------------------------------------------------------------
 * Copyright (c) 2013 - 2020 Arm Limited (or its affiliates). All
 * rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   1.Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   2.Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   3.Neither the name of Arm nor the names of its contributors may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * Name:    DAC_MCB1700.c
 * Purpose: D/A Converter interface for MCB1700 evaluation board
 * Rev.:    1.02
   ---------------------------------------------------------------------------*/

#include "DAC_MCB1700.h"
#include "LPC17xx.h"
#include "PIN_LPC17xx.h"

#define DAC_RESOLUTION        10U        /* Number of D/A converter bits      */

/* DAC pins:
   - AOUT: P0_26
*/

static const PIN DAC_PIN[] = {
  {0U, 26U},
};

/**
  \fn          int32_t DAC_Initialize (void)
  \brief       Initialize Digital-to-Analog Converter
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t DAC_Initialize (void) {

  LPC_SC->PCONP |= (1U << 15);           /* Enable power to IOCON             */

  PIN_Configure (DAC_PIN[0].Portnum, DAC_PIN[0].Pinnum, PIN_FUNC_2, 0U, 0U);
  
  LPC_DAC->DACR  =  0U;                  /* Initialize control register       */

  return 0;
}

/**
  \fn          int32_t DAC_Uninitialize (void)
  \brief       De-initialize Digital-to-Analog Converter
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t DAC_Uninitialize (void) {

  PIN_Configure (DAC_PIN[0].Portnum, DAC_PIN[0].Pinnum, 0U, 0U, 0U);

  return 0;
}

/**
  \fn          int32_t DAC_SetValue (uint32_t val)
  \brief       Set value to be converted
  \param[in]   val  value to convert
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t DAC_SetValue (uint32_t val) {

  LPC_DAC->DACR = (LPC_DAC->DACR & (~(0x3FFU << 6))) | (val & (0x3FFU << 6));

  return 0;
}

/**
  \fn          uint32_t DAC_GetResolution (void)
  \brief       Get resolution of Digital-to-Analog Converter
  \returns     Resolution (in bits)
*/
uint32_t DAC_GetResolution (void) {
  return DAC_RESOLUTION;
}
