/*-----------------------------------------------------------------------------
 * Copyright (c) 2013 - 2020 Arm Limited (or its affiliates). All
 * rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   1.Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   2.Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   3.Neither the name of Arm nor the names of its contributors may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *-----------------------------------------------------------------------------
 * Name:    GLCD_MCB1700.c
 * Purpose: Graphic LCD interface (240x320 pixels) for Graphic LCD with
 *          SPI interface for MCB1700 evaluation board
 * Rev.:    1.0.4
   ---------------------------------------------------------------------------*/

#include "LPC17xx.h"
#include "PIN_LPC17xx.h"
#include "GPIO_LPC17xx.h"
#include "GLCD_Config.h"
#include "Board_GLCD.h"
#include "Driver_SPI.h"

#define UNUSED(x) (void)(x)             /* macro to get rid of 'unused parameter' warning */

/* SPI Interface: SSP1

   PINS:
   - CS     = P0.6 (GPIO pin)
   - RS     = GND
   - WR/SCK = P0.7 (SCK1)
   - RD     = GND
   - SDO    = P0.8 (MISO1)
   - SDI    = P0.9 (MOSI1)                                                    */

/* SPI Driver */
extern ARM_DRIVER_SPI         Driver_SPI1;
#define ptrSPI              (&Driver_SPI1)

/* SPI 1 Pin definitions */
static const PIN spi1_pins[] = {
  { 0U, 6U },
  { 0U, 7U },
  { 0U, 8U },
  { 0U, 9U }
};

#define SPI_PINS_CS      (&spi1_pins[0])
#define SPI_PINS_CLK     (&spi1_pins[1])
#define SPI_PINS_MISO    (&spi1_pins[2])
#define SPI_PINS_MOSI    (&spi1_pins[3])

/*********************** Hardware specific configuration **********************/

#define SPI_START            (0x70U)    /* Start byte for SPI transfer        */
#define SPI_RD               (0x01U)    /* WR bit 1 within start              */
#define SPI_WR               (0x00U)    /* WR bit 0 within start              */
#define SPI_DATA             (0x02U)    /* RS bit 1 within start byte         */
#define SPI_INDEX            (0x00U)    /* RS bit 0 within start byte         */

#define MODE_IN      0x00U
#define MODE_OUT     0x01U

/************************ Global variables ************************************/

static uint16_t   foreground_color = GLCD_COLOR_BLACK;
static uint16_t   background_color = GLCD_COLOR_WHITE;
static GLCD_FONT *active_font      = NULL;
static uint8_t    Himax;
#if (GLCD_SWAP_XY == 0)
static uint32_t   Scroll;
#endif

/************************ Local auxiliary functions ***************************/

/**
  \fn          void delay_ms (uint32_t cnt)
  \brief       Delay in while loop cycles
  \param[in]   cnt   number of while cycles to delay
*/
__STATIC_INLINE void delay_ms (uint32_t cnt) {

  for (cnt = (SystemCoreClock >> 14U) * cnt; cnt > 0U; cnt--) {
    __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP();
    __NOP(); __NOP(); __NOP(); __NOP(); __NOP(); __NOP();
  }
}

/**
  \fn          uint8_t spi_tran_man (uint8_t byte, uint32_t mode)
  \brief       Transfer 1 byte over the serial communication
  \param[in]   byte   byte to be sent
  \param[in]   mode   MODE_OUT = transmit byte, MODE_IN = receive byte
  \returns     byte read while sending
*/
__STATIC_INLINE uint8_t spi_tran_man (uint8_t byte, uint32_t mode) {
  uint8_t val = 0U;
  int32_t i;

  GPIO_SetDir (SPI_PINS_MOSI->Portnum,  SPI_PINS_MOSI->Pinnum,  (mode == MODE_OUT));
  for (i = 7; i >= 0; i--) {
    GPIO_PinWrite (SPI_PINS_CLK->Portnum, SPI_PINS_CLK->Pinnum, 0U);
    delay_ms(1);
    if (mode == MODE_OUT) {
      GPIO_PinWrite (SPI_PINS_MOSI->Portnum, SPI_PINS_MOSI->Pinnum, ((byte & (1U << i)) != 0));
    } else {
      val |= (GPIO_PinRead(SPI_PINS_MOSI->Portnum, SPI_PINS_MOSI->Pinnum) << i);
    }
    GPIO_PinWrite (SPI_PINS_CLK->Portnum, SPI_PINS_CLK->Pinnum, 1U);
    delay_ms(1);
  }

  return (val);
}

/**
  \fn          uint8_t spi_tran (uint8_t byte)
  \brief       Transfer single byte over SPI
  \param[in]   byte  data to send
  \returns     data read
*/
__STATIC_INLINE uint8_t spi_tran (uint8_t byte) {
  uint8_t rcv;

  ptrSPI->Transfer(&byte, &rcv, 1U);
  while (ptrSPI->GetStatus().busy);

  return (rcv);
}

/**
  \fn          void wr_cmd (uint8_t cmd)
  \brief       Write a command to the LCD controller
  \param[in]   cmd  Command to write
*/
__STATIC_INLINE void wr_cmd (uint8_t cmd) {
  uint8_t wr_data[3];

  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  wr_data[0] = SPI_START | SPI_WR | SPI_INDEX;   /* Write : RS = 0, RW = 0 */
  wr_data[1] = 0;
  wr_data[2] = cmd;
  ptrSPI->Send(wr_data, 3U);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);
}

/**
  \fn          void wr_dat (uint16_t dat)
  \brief       Write data to the LCD controller
  \param[in]   dat  Data to write
*/
__STATIC_INLINE void wr_dat (uint16_t dat) {
  uint8_t wr_data[3];

  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  wr_data[0] = SPI_START | SPI_WR | SPI_DATA;    /* RS = 0, RW = 0 */
  wr_data[1] = (dat >>   8);
  wr_data[2] = (dat & 0xFFU);
  ptrSPI->Send(wr_data, 3U);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);
}

/**
  \fn          void wr_dat_start (void)
  \brief       Start writing data to the LCD controller
*/
__STATIC_INLINE void wr_dat_start (void) {

  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  spi_tran(SPI_START | SPI_WR | SPI_DATA);       /* Write : RS = 1, RW = 0 */
}

/**
  \fn          void wr_dat_stop (void)
  \brief       Stop writing data to the LCD controller
*/
__STATIC_INLINE void wr_dat_stop (void) {

  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);
}

/**
  \fn          void wr_dat_only (uint16_t dat)
  \brief       Write data to the LCD controller (without start/stop sequence)
  \param[in]   dat  Data to write
*/
__STATIC_INLINE void wr_dat_only (uint16_t dat) {
  uint8_t wr_data[2];

  wr_data[0] = (dat >>   8);
  wr_data[1] = (dat & 0xFFU);

  ptrSPI->Send(wr_data, 2);
  while (ptrSPI->GetStatus().busy);
}

/**
  \fn          uint16_t rd_dat (void)
  \brief       Read data from the LCD controller
  \returns     Data read
*/
__STATIC_INLINE uint16_t rd_dat (void) {
  uint16_t val = 0U;

  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  spi_tran(SPI_START | SPI_RD | SPI_DATA);       /* Read: RS = 1, RW = 1 */
  spi_tran(0);                                   /* Dummy read 1 */
  val   = spi_tran(0);                           /* Read D8..D15 */
  val <<= 8U;
  val  |= spi_tran(0);                           /* Read D0..D7 */
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);

  return (val);
}

/**
  \fn          void wr_reg (uint16_t reg, uint16_t val)
  \brief       Write a value to the LCD register
  \param[in]   reg  Register to be written
  \param[in]   val  Value to write to the register
*/
__STATIC_INLINE void wr_reg (uint16_t reg, uint16_t val) {
  uint8_t wr_data[3];

  wr_data[0] = SPI_START | SPI_WR | SPI_INDEX;   /* RS = 0, RW = 0 */
  wr_data[1] = (reg >>   8);
  wr_data[2] = (reg & 0xFF);

  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  ptrSPI->Send   (wr_data, 3U);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);

  wr_data[0] = SPI_START | SPI_WR | SPI_DATA;    /* RS = 1, RW = 0 */
  wr_data[1] = (val >>   8);
  wr_data[2] = (val & 0xFFU);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_ACTIVE);
  ptrSPI->Send   (wr_data, 3U);
  while (ptrSPI->GetStatus().busy);
  ptrSPI->Control(ARM_SPI_CONTROL_SS, ARM_SPI_SS_INACTIVE);
}

/**
  \fn          uint16_t rd_reg (uint8_t reg)
  \brief       Read a value from the LCD register
  \param[in]   reg  Register to be read
  \returns     Value read from the register
*/
__STATIC_INLINE uint16_t rd_reg (uint8_t reg) {

  wr_cmd(reg);

  return (rd_dat());
}

/**
  \fn          uint16_t rd_id_man (void)
  \brief       Read LCD controller ID (Himax GLCD)
  \returns     controller ID
*/
__STATIC_INLINE uint16_t rd_id_man (void) {
  uint16_t val;

  /* Set MOSI, MISO and SCK as GPIO pins, with pull-down/pull-up disabled */
  PIN_Configure (SPI_PINS_CS->Portnum,   SPI_PINS_CS->Pinnum,   0U, PIN_PINMODE_TRISTATE, PIN_PINMODE_NORMAL);
  PIN_Configure (SPI_PINS_CLK->Portnum,  SPI_PINS_CLK->Pinnum,  0U, PIN_PINMODE_TRISTATE, PIN_PINMODE_NORMAL);
  PIN_Configure (SPI_PINS_MISO->Portnum, SPI_PINS_MISO->Pinnum, 0U, PIN_PINMODE_TRISTATE, PIN_PINMODE_NORMAL);
  PIN_Configure (SPI_PINS_MOSI->Portnum, SPI_PINS_MOSI->Pinnum, 0U, PIN_PINMODE_TRISTATE, PIN_PINMODE_NORMAL);

  GPIO_SetDir (SPI_PINS_CLK->Portnum,  SPI_PINS_CLK->Pinnum,  GPIO_DIR_OUTPUT);  /* SCK pin is GPIO output */

  GPIO_PinWrite (SPI_PINS_CS->Portnum,  SPI_PINS_CS->Pinnum,  1U);                /* Set chip select high */
  GPIO_PinWrite (SPI_PINS_CLK->Portnum, SPI_PINS_CLK->Pinnum, 1U);                /* Set clock high */

  GPIO_PinWrite (SPI_PINS_CS->Portnum,  SPI_PINS_CS->Pinnum,  0U);
  spi_tran_man (SPI_START | SPI_WR | SPI_INDEX, MODE_OUT);
  spi_tran_man (0x00, MODE_OUT);
  GPIO_PinWrite (SPI_PINS_CS->Portnum,  SPI_PINS_CS->Pinnum,  1U);

  GPIO_PinWrite (SPI_PINS_CS->Portnum,  SPI_PINS_CS->Pinnum,  0U);
  spi_tran_man (SPI_START | SPI_RD | SPI_DATA, MODE_OUT);
  val = spi_tran_man(0, MODE_IN);
  GPIO_PinWrite (SPI_PINS_CS->Portnum,  SPI_PINS_CS->Pinnum,  1U);

  return (val);
}

/**
  \fn          void GLCD_SetWindow (uint32_t x, uint32_t y, uint32_t w, uint32_t h)
  \brief       Set draw window region
  \param[in]   x  Horizontal position in pixel
  \param[in]   y  Vertical position in pixel
  \param[in]   w  Window width in pixel
  \param[in]   h  Window height in pixel
*/
__STATIC_INLINE void GLCD_SetWindow (uint32_t x, uint32_t y, uint32_t w, uint32_t h) {
  uint32_t xe = x + w - 1;
  uint32_t ye = y + h - 1;

  if (Himax) {
    wr_reg(0x02, (uint16_t)(x  >>    8));           /* Column address start MSB */
    wr_reg(0x03, (uint16_t)(x  &  0xFF));           /* Column address start LSB */
    wr_reg(0x04, (uint16_t)(xe >>    8));           /* Column address end MSB */
    wr_reg(0x05, (uint16_t)(xe &  0xFF));           /* Column address end LSB */

    wr_reg(0x06, (uint16_t)(y  >>    8));           /* Row address start MSB */
    wr_reg(0x07, (uint16_t)(y  &  0xFF));           /* Row address start LSB */
    wr_reg(0x08, (uint16_t)(ye >>    8));           /* Row address end MSB */
    wr_reg(0x09, (uint16_t)(ye &  0xFF));           /* Row address end LSB */
  }
  else {
  #if (GLCD_SWAP_XY == 1)
    wr_reg(0x50, (uint16_t)(y ));                   /* Vertical   GRAM Start Address */
    wr_reg(0x51, (uint16_t)(ye));                   /* Vertical   GRAM End   Address */
    wr_reg(0x52, (uint16_t)(x ));                   /* Horizontal GRAM Start Address */
    wr_reg(0x53, (uint16_t)(xe));                   /* Horizontal GRAM End   Address */
    wr_reg(0x20, (uint16_t)(y ));
    wr_reg(0x21, (uint16_t)(x ));
  #else
    wr_reg(0x50, (uint16_t)(x ));                   /* Horizontal GRAM Start Address */
    wr_reg(0x51, (uint16_t)(xe));                   /* Horizontal GRAM End   Address */
    wr_reg(0x52, (uint16_t)(y ));                   /* Vertical   GRAM Start Address */
    wr_reg(0x53, (uint16_t)(ye));                   /* Vertical   GRAM End   Address */
    wr_reg(0x20, (uint16_t)(x ));
    wr_reg(0x21, (uint16_t)(y ));
  #endif
  }
}


/************************ Exported functions **********************************/

/**
  \fn          int32_t GLCD_Initialize (void)
  \brief       Initialize Graphic LCD
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_Initialize (void) {
  uint16_t driverCode = 0;

#if (GLCD_SWAP_XY == 0)
  Scroll = 0;
#endif

  GPIO_PortClock(1U);                    /* Enable GPIO clock */

  /* CS (P0.6) is GPIO output set to high */
  PIN_Configure (0U, 6U, PIN_FUNC_0, PIN_PINMODE_PULLUP, PIN_PINMODE_NORMAL);
  GPIO_SetDir   (0U, 6U, GPIO_DIR_OUTPUT);
  GPIO_PinWrite (0U, 6U, 1U);

  driverCode = rd_id_man();

  ptrSPI->Initialize(NULL);
  ptrSPI->PowerControl(ARM_POWER_FULL);
  ptrSPI->Control(ARM_SPI_MODE_MASTER      |
                  ARM_SPI_CPOL1_CPHA1      |
                  ARM_SPI_MSB_LSB          |
                  ARM_SPI_SS_MASTER_SW     |
                  ARM_SPI_DATA_BITS(8),
                  12500000);

  if (driverCode == 0) {
    driverCode = rd_reg(0x00U);
  }

  if (driverCode == 0x47) {             /* LCD with HX8347-D LCD Controller   */
    Himax = 1;                          /* Set Himax LCD controller flag      */
    /* Driving ability settings ----------------------------------------------*/
    wr_reg(0xEAU, 0x00U);               /* Power control internal used (1)    */
    wr_reg(0xEBU, 0x20U);               /* Power control internal used (2)    */
    wr_reg(0xECU, 0x0CU);               /* Source control internal used (1)   */
    wr_reg(0xEDU, 0xC7U);               /* Source control internal used (2)   */
    wr_reg(0xE8U, 0x38U);               /* Source output period Normal mode   */
    wr_reg(0xE9U, 0x10U);               /* Source output period Idle mode     */
    wr_reg(0xF1U, 0x01U);               /* RGB 18-bit interface ;0x0110       */
    wr_reg(0xF2U, 0x10U);

    /* Adjust the Gamma Curve ------------------------------------------------*/
    wr_reg(0x40U, 0x01U);
    wr_reg(0x41U, 0x00U);
    wr_reg(0x42U, 0x00U);
    wr_reg(0x43U, 0x10U);
    wr_reg(0x44U, 0x0EU);
    wr_reg(0x45U, 0x24U);
    wr_reg(0x46U, 0x04U);
    wr_reg(0x47U, 0x50U);
    wr_reg(0x48U, 0x02U);
    wr_reg(0x49U, 0x13U);
    wr_reg(0x4AU, 0x19U);
    wr_reg(0x4BU, 0x19U);
    wr_reg(0x4CU, 0x16U);

    wr_reg(0x50U, 0x1BU);
    wr_reg(0x51U, 0x31U);
    wr_reg(0x52U, 0x2FU);
    wr_reg(0x53U, 0x3FU);
    wr_reg(0x54U, 0x3FU);
    wr_reg(0x55U, 0x3EU);
    wr_reg(0x56U, 0x2FU);
    wr_reg(0x57U, 0x7BU);
    wr_reg(0x58U, 0x09U);
    wr_reg(0x59U, 0x06U);
    wr_reg(0x5AU, 0x06U);
    wr_reg(0x5BU, 0x0CU);
    wr_reg(0x5CU, 0x1DU);
    wr_reg(0x5DU, 0xCCU);

    /* Power voltage setting -------------------------------------------------*/
    wr_reg(0x1BU, 0x1BU);
    wr_reg(0x1AU, 0x01U);
    wr_reg(0x24U, 0x2FU);
    wr_reg(0x25U, 0x57U);
    wr_reg(0x23U, 0x88U);

    /* Power on setting ------------------------------------------------------*/
    wr_reg(0x18U, 0x36U);               /* Internal oscillator frequency adj  */
    wr_reg(0x19U, 0x01U);               /* Enable internal oscillator         */
    wr_reg(0x01U, 0x00U);               /* Normal mode, no scrool             */
    wr_reg(0x1FU, 0x88U);               /* Power control 6 - DDVDH Off        */
    delay_ms(200U);
    wr_reg(0x1FU, 0x82U);               /* Power control 6 - Step-up: 3 x VCI */
    delay_ms(50U);
    wr_reg(0x1FU, 0x92U);               /* Power control 6 - Step-up: On      */
    delay_ms(50U);
    wr_reg(0x1FU, 0xD2U);               /* Power control 6 - VCOML active     */
    delay_ms(50U);

    /* Color selection -------------------------------------------------------*/
    wr_reg(0x17U, 0x55U);               /* RGB, System interface: 16 Bit/Pixel*/
    wr_reg(0x00U, 0x00U);               /* Scrolling off, no standby          */

    /* Interface config ------------------------------------------------------*/
    wr_reg(0x2FU, 0x11U);               /* LCD Drive: 1-line inversion        */
    wr_reg(0x31U, 0x00U);
    wr_reg(0x32U, 0x00U);               /* DPL=0, HSPL=0, VSPL=0, EPL=0       */

    /* Display on setting ----------------------------------------------------*/
    wr_reg(0x28U, 0x38U);               /* PT(0,0) active, VGL/VGL            */
    delay_ms(200U);
    wr_reg(0x28U, 0x3CU);               /* Display active, VGL/VGL            */

    /* Miror and Swap                                                         */
    wr_reg (0x16U, ((GLCD_MIRROR_Y << 7) |   /* MY                            */
                   (GLCD_MIRROR_X  << 6) |   /* MX                            */
                   (GLCD_SWAP_XY   << 5) |   /* MV                            */
                   (0U             << 4) |   /* ML                            */
                   (1U             << 3)  ));/* BGR                           */

    /* Display scrolling settings --------------------------------------------*/
    wr_reg(0x0EU, 0x00U);               /* TFA MSB                            */
    wr_reg(0x0FU, 0x00U);               /* TFA LSB                            */
    wr_reg(0x10U, 320U >> 8);           /* VSA MSB                            */
    wr_reg(0x11U, 320U &  0xFFU);       /* VSA LSB                            */
    wr_reg(0x12U, 0x00U);               /* BFA MSB                            */
    wr_reg(0x13U, 0x00U);               /* BFA LSB                            */
  }
  else {
    Himax = 0;                          /* This is not Himax LCD controller   */
    /* Start Initial Sequence ------------------------------------------------*/
    /* Miror S1 - S720*/
    wr_reg(0x01U,(!GLCD_MIRROR_X << 8)); /* Set SS bit                        */

    wr_reg(0x02U, 0x0700);              /* Set 1 line inversion               */
    wr_reg(0x04U, 0x0000U);             /* Resize register                    */
    wr_reg(0x08U, 0x0207U);             /* 2 lines front, 7 back porch        */
    wr_reg(0x09U, 0x0000U);             /* Set non-disp area refresh cyc ISC  */
    wr_reg(0x0AU, 0x0000U);             /* FMARK function                     */
    wr_reg(0x0CU, 0x0000U);             /* RGB interface setting              */
    wr_reg(0x0DU, 0x0000U);             /* Frame marker Position              */
    wr_reg(0x0FU, 0x0000U);             /* RGB interface polarity             */

    /* Power On sequence -----------------------------------------------------*/
    wr_reg(0x10U, 0x0000U);             /* Reset Power Control 1              */
    wr_reg(0x11U, 0x0000U);             /* Reset Power Control 2              */
    wr_reg(0x12U, 0x0000U);             /* Reset Power Control 3              */
    wr_reg(0x13U, 0x0000U);             /* Reset Power Control 4              */
    delay_ms(200U);                     /* Discharge cap power voltage (200ms)*/
    wr_reg(0x10U, 0x12B0U);             /* SAP, BT[3:0], AP, DSTB, SLP, STB   */
    wr_reg(0x11U, 0x0007U);             /* DC1[2:0], DC0[2:0], VC[2:0]        */
    delay_ms(50U);                      /* Delay 50 ms                        */
    wr_reg(0x12U, 0x01BDU);             /* VREG1OUT voltage                   */
    delay_ms(50U);                      /* Delay 50 ms                        */
    wr_reg(0x13U, 0x1400U);             /* VDV[4:0] for VCOM amplitude        */
    wr_reg(0x29U, 0x000EU);             /* VCM[4:0] for VCOMH                 */
    delay_ms(50U);                      /* Delay 50 ms                        */
    wr_reg(0x20U, 0x0000U);             /* GRAM horizontal Address            */
    wr_reg(0x21U, 0x0000U);             /* GRAM Vertical Address              */

    /* Adjust the Gamma Curve ------------------------------------------------*/
    switch (driverCode) {
      case 0x5408U:                     /* LCD with SPFD5408 LCD Controller   */
        wr_reg(0x30U, 0x0B0DU);
        wr_reg(0x31U, 0x1923U);
        wr_reg(0x32U, 0x1C26U);
        wr_reg(0x33U, 0x261CU);
        wr_reg(0x34U, 0x2419U);
        wr_reg(0x35U, 0x0D0BU);
        wr_reg(0x36U, 0x1006U);
        wr_reg(0x37U, 0x0610U);
        wr_reg(0x38U, 0x0706U);
        wr_reg(0x39U, 0x0304U);
        wr_reg(0x3AU, 0x0E05U);
        wr_reg(0x3BU, 0x0E01U);
        wr_reg(0x3CU, 0x010EU);
        wr_reg(0x3DU, 0x050EU);
        wr_reg(0x3EU, 0x0403U);
        wr_reg(0x3FU, 0x0607U);
        break;

      case 0x2593/*0x9325U*/:           /* LCD with RM68050 LCD Controller    */
        wr_reg(0x0030U,0x0000U);
        wr_reg(0x0031U,0x0607U);
        wr_reg(0x0032U,0x0305U);
        wr_reg(0x0035U,0x0000U);
        wr_reg(0x0036U,0x1604U);
        wr_reg(0x0037U,0x0204U);
        wr_reg(0x0038U,0x0001U);
        wr_reg(0x0039U,0x0707U);
        wr_reg(0x003CU,0x0000U);
        wr_reg(0x003DU,0x000FU);
        break;

      case 0x9320U:                     /* LCD with ILI9320 LCD Controller    */
      default:                          /* LCD with other LCD Controller      */
        wr_reg(0x30U, 0x0006U);
        wr_reg(0x31U, 0x0101U);
        wr_reg(0x32U, 0x0003U);
        wr_reg(0x35U, 0x0106U);
        wr_reg(0x36U, 0x0B02U);
        wr_reg(0x37U, 0x0302U);
        wr_reg(0x38U, 0x0707U);
        wr_reg(0x39U, 0x0007U);
        wr_reg(0x3CU, 0x0600U);
        wr_reg(0x3DU, 0x020BU);
        break;
    }

    /* Set GRAM area ---------------------------------------------------------*/
    wr_reg(0x50U, 0x0000U);             /* Horizontal GRAM Start Address      */
    wr_reg(0x51U, (GLCD_HEIGHT - 1));   /* Horizontal GRAM End   Address      */
    wr_reg(0x52U, 0x0000U);             /* Vertical   GRAM Start Address      */
    wr_reg(0x53U, (GLCD_WIDTH - 1));    /* Vertical   GRAM End   Address      */

    /* Set Gate Scan Line ----------------------------------------------------*/
    switch (driverCode) {
      case 0x5408U:                     /* LCD with SPFD5408 LCD Controller   */
      case 0x2593U/*0x9325*/:           /* LCD with RM68050 LCD Controller    */
        wr_reg(0x60U, (0x2700U | ((!GLCD_MIRROR_Y) << 15)));
        break;

      case 0x9320U:                     /* LCD with ILI9320 LCD Controller    */
      default:                          /* LCD with other LCD Controller      */
        wr_reg(0x60U, (0x2700U | (GLCD_MIRROR_Y << 15)));
        break;
    }
    wr_reg(0x61U, 0x0001U);             /* NDL,VLE, REV                       */
    wr_reg(0x6AU, 0x0000U);             /* Set scrolling line                 */

    /* Partial Display Control -----------------------------------------------*/
    wr_reg(0x80U, 0x0000U);
    wr_reg(0x81U, 0x0000U);
    wr_reg(0x82U, 0x0000U);
    wr_reg(0x83U, 0x0000U);
    wr_reg(0x84U, 0x0000U);
    wr_reg(0x85U, 0x0000U);

    /* Panel Control ---------------------------------------------------------*/
    wr_reg(0x90U, 0x0010U);
    wr_reg(0x92U, 0x0000U);
    wr_reg(0x93U, 0x0003U);
    wr_reg(0x95U, 0x0110U);
    wr_reg(0x97U, 0x0000U);
    wr_reg(0x98U, 0x0000U);

    /* Set GRAM write direction
            I/D=11,BGR=1,  AM */
    wr_reg(0x03U, (0x1030U | (GLCD_SWAP_XY << 3)));

    wr_reg(0x07U, 0x0137U);             /* 262K color and display ON */
  }

  GPIO_PinWrite (4U, 28U, 1U);          /* Turn LCD Backlight On */

  return 0;
}

/**
  \fn          int32_t GLCD_Uninitialize (void)
  \brief       De-initialize Graphic LCD
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_Uninitialize (void) {

  ptrSPI->PowerControl(ARM_POWER_OFF);  /* Turn off SPI*/
  ptrSPI->Uninitialize();               /* uninitialize SPI*/
  GPIO_PinWrite (4U, 28U, 0U);          /* Turn backlight off */

  return 0;
}

/**
  \fn          int32_t GLCD_SetForegroundColor (uint32_t color)
  \brief       Set foreground color
  \param[in]   color  Color value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetForegroundColor (uint32_t color) {
  foreground_color = color & 0xFFFFU;
  return 0;
}

/**
  \fn          int32_t GLCD_SetBackgroundColor (uint32_t color)
  \brief       Set background color
  \param[in]   color  Color value
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetBackgroundColor (uint32_t color) {
  background_color = color & 0xFFFFU;
  return 0;
}

/**
  \fn          int32_t GLCD_ClearScreen (void)
  \brief       Clear screen (with active background color)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_ClearScreen (void) {
  uint32_t i;

  GLCD_SetWindow(0U, 0U, GLCD_WIDTH, GLCD_HEIGHT);

  wr_cmd(0x22U);
  wr_dat_start();
  for (i = 0U; i < (GLCD_WIDTH * GLCD_HEIGHT); i++) {
    wr_dat_only(background_color);
  }
  wr_dat_stop();

  return 0;
}

/**
  \fn          int32_t GLCD_SetFont (GLCD_FONT *font)
  \brief       Set active font
  \param[in]   font   Pointer to font structure
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_SetFont (GLCD_FONT *font) {
  active_font = font;
  return 0;
}

/**
  \fn          int32_t GLCD_DrawPixel (uint32_t x, uint32_t y)
  \brief       Draw pixel (in active foreground color)
  \param[in]   x  x position in pixels (0 = left corner)
  \param[in]   y  y position in pixels (0 = upper corner)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawPixel (uint32_t x, uint32_t y) {

#if (GLCD_SWAP_XY == 0)
  y = (y + Scroll) % GLCD_HEIGHT;
#endif

  GLCD_SetWindow(0U, 0U, GLCD_WIDTH, GLCD_HEIGHT);

  if (Himax) {
    wr_reg(0x02U, (uint16_t)(x >>    8));
    wr_reg(0x03U, (uint16_t)(x &  0xFF));
    wr_reg(0x04U, (uint16_t)(x >>    8));
    wr_reg(0x05U, (uint16_t)(x &  0xFF));

    wr_reg(0x06U, (uint16_t)(y >>    8));
    wr_reg(0x07U, (uint16_t)(y &  0xFF));
    wr_reg(0x08U, (uint16_t)(y >>    8));
    wr_reg(0x09U, (uint16_t)(y &  0xFF));
  }
  else {
  #if (GLCD_SWAP_XY == 1)
    wr_reg(0x20U, (uint16_t)(y));
    wr_reg(0x21U, (uint16_t)(x));
  #else
    wr_reg(0x20U, (uint16_t)(x));
    wr_reg(0x21U, (uint16_t)(y));
  #endif
  }

  wr_cmd(0x22U);
  wr_dat(foreground_color);

  return 0;
}

/**
  \fn          int32_t GLCD_DrawHLine (uint32_t x, uint32_t y, uint32_t length)
  \brief       Draw horizontal line (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   length Line length
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawHLine (uint32_t x, uint32_t y, uint32_t length) {

#if (GLCD_SWAP_XY == 0)
  y = (y + Scroll) % GLCD_HEIGHT;
#endif

  GLCD_SetWindow(x, y, length, 1);
  wr_cmd(0x22U);
  wr_dat_start();

  while (length--) {
    wr_dat_only(foreground_color);
  }
  wr_dat_stop();

  return 0;
}

/**
  \fn          int32_t GLCD_DrawVLine (uint32_t x, uint32_t y, uint32_t length)
  \brief       Draw vertical line (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   length Line length in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawVLine (uint32_t x, uint32_t y, uint32_t length) {

#if (GLCD_SWAP_XY == 0)
  y = (y + Scroll) % GLCD_HEIGHT;
#endif

  GLCD_SetWindow(x, y, 1U, length);
  wr_cmd(0x22U);
  wr_dat_start();

  while (length--) {
    wr_dat_only(foreground_color);
  }
  wr_dat_stop();

  return 0;
}

/**
  \fn          int32_t GLCD_DrawRectangle (uint32_t x, uint32_t y, uint32_t width, uint32_t height)
  \brief       Draw rectangle (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Rectangle width in pixels
  \param[in]   height Rectangle height in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawRectangle (uint32_t x, uint32_t y, uint32_t width, uint32_t height) {

  GLCD_DrawHLine (x,         y,          width);
  GLCD_DrawHLine (x,         y + height, width);
  GLCD_DrawVLine (x,         y,          height);
  GLCD_DrawVLine (x + width, y,          height);

  return 0;
}

/**
  \fn          int32_t GLCD_DrawChar (uint32_t x, uint32_t y, int32_t ch)
  \brief       Draw character (in active foreground color)
  \param[in]   x      x position in pixels (0 = left corner)
  \param[in]   y      y position in pixels (0 = upper corner)
  \param[in]   ch     Character
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawChar (uint32_t x, uint32_t y, int32_t ch) {
  uint32_t i, j;
  uint32_t wb;
  uint8_t *ptr_ch_bmp;

#if (GLCD_SWAP_XY == 0)
  y = (y + Scroll) % GLCD_HEIGHT;
#endif

  if (active_font == NULL) return -1;

  ch        -= active_font->offset;
  wb         = (active_font->width + 7)/8;
  ptr_ch_bmp = (uint8_t *)(uint32_t)active_font->bitmap + ((uint32_t)ch * wb * active_font->height);

  GLCD_SetWindow(x, y, active_font->width, active_font->height);

  wr_cmd(0x22U);
  wr_dat_start();

  for (i = 0U; i < active_font->height; i++) {
    for (j = 0U; j < active_font->width; j++) {
      wr_dat_only(((*ptr_ch_bmp >> (j & 7)) & 1) ? foreground_color : background_color);
      if (((j & 7) == 7) && (j != (active_font->width - 1))) ptr_ch_bmp++;
    }
    ptr_ch_bmp++;
  }
  wr_dat_stop();

  return 0;
}

/**
  \fn          int32_t GLCD_DrawString (uint32_t x, uint32_t y, const char *str)
  \brief       Draw string (in active foreground color)
  \param[in]   x      x position in pixels (0 = left corner)
  \param[in]   y      y position in pixels (0 = upper corner)
  \param[in]   str    Null-terminated String
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawString (uint32_t x, uint32_t y, const char *str) {

  while (*str) { GLCD_DrawChar(x, y, *str++); x += active_font->width; }

  return 0;
}

/**
  \fn          int32_t GLCD_DrawBargraph (uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint32_t val)
  \brief       Draw bargraph (in active foreground color)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Full bargraph width in pixels
  \param[in]   height Full bargraph height in pixels
  \param[in]   val    Active bargraph value (maximum value is 100)
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawBargraph (uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint32_t val) {
  uint32_t i, j;

  val = (val * width) / 100;            /* Scale value */

#if (GLCD_SWAP_XY == 0)
  y = (y + Scroll) % GLCD_HEIGHT;
#endif

  GLCD_SetWindow(x, y, width, height);

  wr_cmd(0x22U);
  wr_dat_start();
  for (i = 0U; i < height; i++) {
    for (j = 0U; j < width; j++) {
      wr_dat_only((j >= val) ? background_color : foreground_color);
    }
  }
  wr_dat_stop();

  return 0;
}

/**
  \fn          int32_t GLCD_DrawBitmap (uint32_t x, uint32_t y, uint32_t width, uint32_t height, const uint8_t *bitmap)
  \brief       Draw bitmap (bitmap from BMP file without header)
  \param[in]   x      Start x position in pixels (0 = left corner)
  \param[in]   y      Start y position in pixels (0 = upper corner)
  \param[in]   width  Bitmap width in pixels
  \param[in]   height Bitmap height in pixels
  \param[in]   bitmap Bitmap data
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_DrawBitmap (uint32_t x, uint32_t y, uint32_t width, uint32_t height, const uint8_t *bitmap) {
  uint32_t  i, j;
  uint16_t *ptr_bmp;

  ptr_bmp = (uint16_t *)(uint32_t)bitmap + (width * (height - 1));

#if (GLCD_SWAP_XY == 0)
  y = (y + Scroll) % GLCD_HEIGHT;
#endif

  GLCD_SetWindow(x, y, width, height);

  wr_cmd(0x22U);
  wr_dat_start();

  for (i = 0U; i < height; i++) {
    for (j = 0U; j < width; j++) {
      wr_dat_only((*ptr_bmp++));
    }
    ptr_bmp -= 2*width;
  }
  wr_dat_stop();

  return 0;
}

/**
  \fn          int32_t GLCD_VScroll (uint32_t dy)
  \brief       Scroll picture on display vertically
  \param[in]   dy  Scroll size in pixels
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t GLCD_VScroll (uint32_t dy) {
#if (GLCD_SWAP_XY == 0)
  uint32_t i;

  Scroll = (Scroll + dy) % GLCD_HEIGHT;

  if (Himax) {
    wr_reg(0x01U, 0x08U);
    wr_reg(0x14U, (uint16_t)Scroll >> 8);  /* VSP MSB */
    wr_reg(0x15U, Scroll & 0xFF);          /* VSP LSB */
  }
  else {
    wr_reg(0x6AU, (uint16_t)Scroll);
    wr_reg(0x61U, 3U);
  }

  /* Clear new space */
  GLCD_SetWindow(0U, Scroll, GLCD_WIDTH, dy);

  wr_cmd(0x22U);
  wr_dat_start();
  for(i = 0U; i < (GLCD_WIDTH * dy); i++) {
    wr_dat_only(background_color);
  }
  wr_dat_stop();

  return 0;
#else
  UNUSED(dy);

  return -1;
#endif
}

/**
  \fn          int32_t GLCD_FrameBufferAccess (bool enable)
  \brief       Enable or disable direct access to FrameBuffer
  \param[in]   enable Enable or disable request
                - true (1): enable
                - false(0): disable
  \returns
   - \b  0: function succeeded
   - \b -1: function failed
*/
int32_t  GLCD_FrameBufferAccess (bool enable) {
  UNUSED(enable);

  GLCD_SetWindow(0U, 0U, GLCD_WIDTH, GLCD_HEIGHT);

  wr_reg (0x20U, 0x0000U);
  wr_reg (0x21U, 0x0000U);
  wr_cmd (0x22U);

  return 0;
}

/**
  \fn          uint32_t GLCD_FrameBufferAddress (void)
  \brief       Retrieve FrameBuffer address
  \returns     FrameBuffer address
*/
uint32_t GLCD_FrameBufferAddress (void) {

  return 0;
}
